# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.models.api.__init__
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This package contains the class adaptors of the API logic within the Mapillary Python SDK.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

from . import entities  # noqa: F401
from . import vector_tiles  # noqa: F401
